﻿#
#Depends on Microsoft SharePoint Online Client Components to be installed, these are used even for an On-Premise SharePoint environment
#https://www.microsoft.com/en-us/download/details.aspx?id=42038
# Example usage
# This script requires PowerShell 2.0 and the SharePoint CSOM library
# .\DeployFilesOnPremise.ps1 -SiteUrl "https://your-domain.local/sites/pwa" -Version "2019"
# .\DeployFilesOnPremise.ps1 -SiteUrl "http://project2016.thorappsdev.local/PWA2" -Version "2016"

Param(
    [string]$SiteUrl,
    [string]$Version
)

# Load SharePoint CSOM Assemblies
Add-Type -Path "C:\Program Files\Common Files\microsoft shared\Web Server Extensions\16\ISAPI\Microsoft.SharePoint.Client.dll"
Add-Type -Path "C:\Program Files\Common Files\microsoft shared\Web Server Extensions\16\ISAPI\Microsoft.SharePoint.Client.Runtime.dll"
Add-Type -Path "C:\Program Files\Common Files\microsoft shared\Web Server Extensions\16\ISAPI\Microsoft.SharePoint.Client.Publishing.dll"

#Add-Type -Path "Microsoft.SharePoint.Client.dll"
#Add-Type -Path "Microsoft.SharePoint.Client.Runtime.dll"
#Add-Type -Path "Microsoft.SharePoint.Client.Publishing.dll"

Add-Type -Path "Microsoft.IdentityModel.Clients.ActiveDirectory.dll"

$scriptDir = Split-Path -Path $MyInvocation.MyCommand.Definition -Parent

# File paths relative to the script location
$Files = @(
    "Index.html",
    "LicenseAgreement.aspx",
    "EntitySelection.aspx",
    "Images/AppLogo.png",
    "Styles/styles.css",
    "Scripts/EntitySelection.js",
    "Scripts/InitPage.js",
    "Scripts/IOBaseEntity.js",
    "Scripts/IOLookupTables.js",
    "Scripts/IOLookupEntries.js",
    "Scripts/IOCustomFields.js",
    "Scripts/IOProjectDetailPages.js",
    "Scripts/IOEnterpriseProjectTypes.js",
    "Scripts/IOPWAViews.js",
    "Scripts/IOResources.js"
    "Scripts/IOSiteColumns.js",
    "Scripts/IOContentTypes.js"
)

$LibraryName = "ThorAppsPCM"
$PWASitePath = $SiteUrl -replace 'https?://[^/]+'

function Get-ClientContext($url, $version) {
    $context = New-Object Microsoft.SharePoint.Client.ClientContext($url)

    $creds = [System.Net.CredentialCache]::DefaultCredentials
    #$creds = Get-Credential
    $context.Credentials = $creds

    #Test the connection
    $web = $context.Web
    $context.Load($web)
    $context.ExecuteQuery()
    
    return $context
}

function Ensure-List($context, $listTitle) {
    $lists = $context.Web.Lists
    $context.Load($lists)
    $context.ExecuteQuery()

    $list = $lists | Where-Object { $_.Title -eq $listTitle }

    if ($null -eq $list) {
        $listCreationInfo = New-Object Microsoft.SharePoint.Client.ListCreationInformation
        $listCreationInfo.Title = $listTitle
        $listCreationInfo.TemplateType = [Microsoft.SharePoint.Client.ListTemplateType]::DocumentLibrary
        $list = $context.Web.Lists.Add($listCreationInfo)
        $context.Load($list)
        $context.ExecuteQuery()
    }

    return $list
}

function Add-ContentTypeToList($context, $listName, $contentTypeName) {
    # Load the web and its content types
    $web = $context.Web
    $context.Load($web)
    $context.Load($web.ContentTypes)
    $context.ExecuteQuery()

    # Get the desired content type
    $contentType = $web.ContentTypes | Where-Object { $_.Name -eq $contentTypeName }

    if ($contentType -ne $null) {
        # Load the list and its content types
        $list = $web.Lists.GetByTitle($listName)
        $context.Load($list)
        $context.Load($list.ContentTypes)
        $context.ExecuteQuery()

        # Check if the content type is already added
        $listContentType = $list.ContentTypes | Where-Object { $_.Name -eq $contentTypeName }
        if ($listContentType -eq $null) {
            # Add the content type to the list
            $out = $list.ContentTypes.AddExistingContentType($contentType)
            $context.ExecuteQuery()
            Write-Host "Content type '$contentTypeName' added to list '$listName'."
        } else {
            Write-Host "Content type '$contentTypeName' is already added to list '$listName'."
        }
    } else {
        Write-Host "Content type '$contentTypeName' not found."
    }
}


function Upload-FilesToLibrary($context, $files, $libraryName) {
    
    foreach ($file in $files) {
        $fullFileName = $scriptDir + "\" +$libraryName + "\" + $file.replace('/','\')
        write-host $fullFileName
        $fileContent = [System.IO.File]::ReadAllBytes($fullFileName)

        $folderPath = [System.IO.Path]::GetDirectoryName($file)
        $fileName = [System.IO.Path]::GetFileName($file)

        $fileCreationInfo = New-Object Microsoft.SharePoint.Client.FileCreationInformation
        $fileCreationInfo.Content = $fileContent
        $fileCreationInfo.Url = $fileName
        $fileCreationInfo.Overwrite = $true

        $libraryRootFolder = $context.Web.Lists.GetByTitle($libraryName).RootFolder
        $context.Load($libraryRootFolder)

        if ($folderPath -ne '') {
            $folder = $libraryRootFolder.Folders.Add($folderPath)
            $context.Load($folder)
            $context.ExecuteQuery()
            $uploadedFile = $folder.Files.Add($fileCreationInfo)
        } else {
            $uploadedFile = $libraryRootFolder.Files.Add($fileCreationInfo)
        }

        $context.Load($uploadedFile)
        $context.ExecuteQuery()
    }
}

function Add-WebPartToClassicPage($context, $pageUrl, $webPartXml, $zoneId, $zoneIndex) {
    Write-Host "AddClassic Web Part to PageUrl=$pageUrl"
    $file = $context.Web.GetFileByServerRelativeUrl($pageUrl)
    $webPartManager = $file.GetLimitedWebPartManager([Microsoft.SharePoint.Client.WebParts.PersonalizationScope]::Shared)
    $context.Load($webPartManager)

    # Check if the web part already exists on the page
    $existingWebParts = $webPartManager.WebParts
    $context.Load($existingWebParts)
    $context.ExecuteQuery()

    $webPartsToRemove = @()

    ForEach($WebPart in $existingWebParts)
    {
        #Get Web part properties
        $context.Load($Webpart.WebPart.Properties)
        $context.ExecuteQuery()           
        If($Webpart.WebPart.Properties.FieldValues["Title"] -eq 'PWA Configuration Manager')
        {
            $webPartsToRemove.Add($Webpart.ID)
        }
    }

    #$webPartsToRemove = $existingWebParts # | Where-Object { $_.WebPart.Title -eq "PWA Configuration Manager" }

    Write-Host "Removing existing web parts with title 'PWA Configuration Manager'..."
    foreach ($webPartId in $webPartsToRemove) {
        #Remove-PnPWebPart -ServerRelativePageUrl $pageUrl -Identity $webPart.Id

        $webpartToRemove=$webPartManager.WebParts.GetById($webPartId)
        $webpartToRemove.DeleteWebPart()
        $context.ExecuteQuery()
    }
    $context.ExecuteQuery()
    Write-Host "Existing web parts removed successfully."

    $importWebPart = $webPartManager.ImportWebPart($webPartXml)
    $webPart = $webPartManager.AddWebPart($importWebPart.WebPart, $zoneId, $zoneIndex)
    $context.Load($webPart)
    $context.ExecuteQuery()
}

function Rename-FileInLibrary {
    param (
        [Parameter(Mandatory = $true)]
        [Microsoft.SharePoint.Client.ClientContext]$context,

        [Parameter(Mandatory = $true)]
        [string]$libraryName,

        [Parameter(Mandatory = $true)]
        [string]$currentFileName,

        [Parameter(Mandatory = $true)]
        [string]$newFileName
    )

    $library = $context.Web.Lists.GetByTitle($libraryName)
    $context.Load($library)
    $context.ExecuteQuery()

    $file = $library.RootFolder.Files.GetByUrl($library.RootFolder.ServerRelativeUrl + "/" + $currentFileName)
    $context.Load($file)
    $context.ExecuteQuery()

    $file.MoveTo($library.RootFolder.ServerRelativeUrl + "/" + $newFileName, [Microsoft.SharePoint.Client.MoveOperations]::Overwrite)
    $context.ExecuteQuery()

    Write-Host "File renamed successfully: $($file.ServerRelativeUrl)"
}


# Get Client Context
$context = Get-ClientContext -url $SiteUrl -version $Version

# Ensure Library
$library = Ensure-List -context $context -listTitle $LibraryName

Add-ContentTypeToList -context $context -listName $LibraryName -contentTypeName "Web Part Page"

# Upload Files
Upload-FilesToLibrary -context $context -files $Files -libraryName $LibraryName



# Add the web part pages
$page = $null

$targetPageName = "PWAGovernance.aspx"
$pageUrl = "$PWASitePath/$LibraryName/" + $targetPageName

try
{
    $page = $context.Web.GetFileByServerRelativeUrl($pageUrl)
    $context.Load($page)
    $context.ExecuteQuery()
}
catch {
    #if ($_.Exception.GetType().FullName -eq "System.Management.Automation.MethodInvocationException") {
        $page = $null
        Write-Host "File not found: $pageUrl"
    #} else {
    #    Write-Host "Other exception: $($_.Exception.GetType().FullName)"
    #}
}

#if ($null -ne $page) {
#    Write-Host "The page $pageUrl already exists. Skipping upload and configuration."
#} else {
 #   Write-Host "The page $pageUrl does not exist. Creating it."
    $templateFileName = "TemplateWebPartPage2016.aspx"
    if ($Version.ToLower() -eq "2019")
    {
        $templateFileName = "TemplateWebPartPage2019.aspx"
    }
    if ($Version.ToLower() -eq "online")
    {
        $templateFileName = "TemplateWebPartPage.aspx"
    }

    $webPartPageFiles = @($templateFileName)

    $templateFilePath = $scriptDir +"\$LibraryName\" + $templateFileName
    write-host "Uploading ...$templateFilePath"

    Upload-FilesToLibrary -context $context -files $webPartPageFiles -libraryName $LibraryName
    Rename-FileInLibrary -context $context -libraryName $LibraryName -currentFileName $templateFileName -newFileName $targetPageName

    $webPartXml = ""

    #if ($version.ToLower() -eq "2016") {
        $webPartXml = @"
<?xml version="1.0" encoding="utf-8"?>
<WebPart xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://schemas.microsoft.com/WebPart/v2">
  <Title>PWA Configuration Manager</Title>
  <FrameType>None</FrameType>
  <Description>Web Part for managing PWA configuration settings.</Description>
  <IsIncluded>true</IsIncluded>
  <ZoneID>Header</ZoneID>
  <PartOrder>0</PartOrder>
  <FrameState>Normal</FrameState>
  <Height />
  <Width />
  <AllowRemove>true</AllowRemove>
  <AllowZoneChange>true</AllowZoneChange>
  <AllowMinimize>true</AllowMinimize>
  <AllowConnect>true</AllowConnect>
  <AllowEdit>true</AllowEdit>
  <AllowHide>true</AllowHide>
  <IsVisible>true</IsVisible>
  <DetailLink />
  <HelpLink />
  <HelpMode>Modeless</HelpMode>
  <Dir>Default</Dir>
  <PartImageSmall />
  <MissingAssembly>Cannot import this Web Part.</MissingAssembly>
  <PartImageLarge>/_layouts/15/images/mscontl.gif</PartImageLarge>
  <IsIncludedFilter />
  <Assembly>Microsoft.SharePoint, Version=16.0.0.0, Culture=neutral, PublicKeyToken=71e9bce111e9429c</Assembly>
  <TypeName>Microsoft.SharePoint.WebPartPages.ContentEditorWebPart</TypeName>
  <ContentLink xmlns="http://schemas.microsoft.com/WebPart/v2/ContentEditor">$PWASitePath/$LibraryName/Index.html</ContentLink>
  <Content xmlns="http://schemas.microsoft.com/WebPart/v2/ContentEditor" />
  <PartStorage xmlns="http://schemas.microsoft.com/WebPart/v2/ContentEditor" />
</WebPart>
"@
    #}

    Add-WebPartToClassicPage -context $context -pageUrl $pageUrl -webPartXml $webPartXml -zoneId "Header" -zoneIndex 0
    Write-Host "The page $pageUrl has been uploaded and configured with the Content Editor Web Part."
#}

# Dispose the Client Context
$context.Dispose()
Get-Date