// Ensure the global SP and PS variables are defined
/// <reference path="PS.debug.js" />
/// <reference path="SP.debug.js" />
/// <reference path="SP.Runtime.js" />
var SP = SP || {};
var PS = PS || {};

class IOContentTypes extends IOBaseEntity {
    constructor(apiUrl, fileInput, logTextArea, progressBar) {
        super(apiUrl, fileInput, logTextArea, progressBar);
    }

    get_name() {
        return "Site Content Types";
    }

    async export() {
        this.startNewLog("Retrieving Content Types");
        const data = await this.request('GET', '/_api/web/contenttypes');
        //const contentTypes = data.value;
        const contentTypes = data.d.results;

        this.appendLog(`Found ${contentTypes.length} Content Types in the Site Collection`);

        let projectContext = PS.ProjectContext.get_current();
        var existingContentTypes = projectContext.get_web().get_contentTypes();
        projectContext.load(existingContentTypes, 'Include(Id, Name, Parent, StringId)');
        await this.executeQuery(projectContext, () => {
            //All good
        },
            (sender, args) => {
                throw new Error(`FAILED to read Content Types from Site Collection'. Error: '${args.get_message()}'.`);
            });

        for (const contentType of contentTypes) {
            var existingContentType = this.selectContentTypeByStringId(existingContentTypes, contentType.StringId);
            var parentContentType = existingContentType.get_parent();
            contentType.ParentContentTypeName = parentContentType.get_name();
            contentType.ParentContentTypeStringId = parentContentType.get_stringId();
        }

        var pwaConfig = {
            ContentTypes: contentTypes
        };

        const blob = new Blob([JSON.stringify(pwaConfig, null, 4)], { type: "application/json" });
        const url = URL.createObjectURL(blob);
        const link = document.createElement('a');
        link.href = url;
        link.download = 'ContentTypes.json';
        link.click();
    }

    async import() {
        //return new Promise(async (resolve, reject) => {
        //try {
        this.startNewLog('');

        const file = this.openFile();
        this.appendLog("Reading Content Types from file");
        const fileContent = await this.readFileAsText(file);
        const pwaConfig = JSON.parse(fileContent);
        const contentTypes = pwaConfig.ContentTypes;
        if (contentTypes == null) {
            reject(`No '${this.get_name()}' were found in the selected file.`);
        }

        var count = contentTypes.length;
        this.appendLog(`Found ${count} Content Types in file`);

        this.appendLog("Reading Content Types from Site Collection");

        let projectContext = PS.ProjectContext.get_current();
        let existingContentTypes = projectContext.get_web().get_contentTypes();
        let siteColumns = projectContext.get_web().get_fields();

        projectContext.load(siteColumns);
        projectContext.load(existingContentTypes, 'Include(Id, Name, StringId, SchemaXml)');

        await this.executeQuery(projectContext, () => {
            //Success
        },
            (sender, args) => {
                throw new Error(`FAILED to load Content Types. Error: '${args.get_message()}'.`);
            });

        this.appendLog(`Found ${this.count(existingContentTypes)} Content Types in Site Collection`);

        //let siteColumns = projectContext.get_web().get_fields();

        //await this.executeQuery(projectContext, () => {
        //    //Success
        //},
        //(sender, args) => {
        //    reject(`FAILED to load Content Types. Error: '${args.get_message()}'.`);
        //});

        var index = 0;
        var progress = 0;
        for (const contentType of contentTypes) {

            if (!(contentType.Selected == null || contentType.Selected == true)) {
                this.appendLog(`Content Type '${contentType.Name}' is not selected for Import.`);
            }
            else {
                try {
                    if (contentType.Sealed) {
                        this.appendWarning(`Content Type '${contentType.Name}' is marked as "Sealed" and cannot be Imported.`);
                    }
                    else {
                        let existingContentType = this.selectContentTypeByStringId(existingContentTypes, contentType.StringId);
                        let parser = new DOMParser();
                        let xmlDoc = parser.parseFromString(contentType.SchemaXml, "text/xml");

                        if (existingContentType == null) {

                            // Create new content type
                            this.appendLog(`Content Type '${contentType.Name}' does not exist. Creating it.`);


                            //Can only create new Content Type with either a specific ID or with a parent, but not set both. It errors.
                            //let parentContentType = this.selectContentTypeByStringId(types, contentType.ParentContentTypeStringId);

                            let contentTypeName = xmlDoc.getElementsByTagName("ContentType")[0].getAttribute('Name');
                            let contentTypeId = xmlDoc.getElementsByTagName("ContentType")[0].getAttribute('ID');
                            let contentTypeGroup = xmlDoc.getElementsByTagName("ContentType")[0].getAttribute('Group');
                            let contentTypeDescription = xmlDoc.getElementsByTagName("ContentType")[0].getAttribute('Description');

                            // Create a ContentTypeCreationInformation instance
                            let ctCreationInfo = new SP.ContentTypeCreationInformation();
                            ctCreationInfo.set_id(contentTypeId);
                            ctCreationInfo.set_name(contentTypeName);
                            ctCreationInfo.set_description(contentTypeDescription);
                            ctCreationInfo.set_group(contentTypeGroup);
                            //ctCreationInfo.set_parentContentType(parentContentType);

                            // Add the ContentTypeCreationInformation instance to the content types collection
                            let newContentType = existingContentTypes.add(ctCreationInfo);

                            await this.executeQuery(projectContext, () => {
                                this.appendLog(`Content Type '${contentType.Name}' has been created.`);
                            },
                                (sender, args) => {
                                    throw new Error(`FAILED to create Content Type '${contentType.Name}'. Error: '${args.get_message()}'.`);
                                });

                            //Reload Content Types
                            existingContentTypes = projectContext.get_web().get_contentTypes();
                            projectContext.load(existingContentTypes, 'Include(Id, Name, StringId, SchemaXml)');

                            await this.executeQuery(projectContext, () => {
                                //Success
                            },
                                (sender, args) => {
                                    throw new Error(`FAILED to load Content Types. Error: '${args.get_message()}'.`);
                                });
                        }
                        else {
                            // Update existing content type

                            if (existingContentType.get_schemaXml() === contentType.SchemaXml) {
                                //No need to do updates
                                continue; //move on to next content type
                            }
                            //No need for log, just makes it messy
                            //this.appendLog(`Content Type '${contentType.Name}' already exists. Checking for updates.`);
                        }

                        //Refetch ContentType
                        existingContentType = this.selectContentTypeByStringId(existingContentTypes, contentType.StringId);

                        let existingContentTypeFields = existingContentType.get_fields();
                        projectContext.load(existingContentTypeFields);
                        await this.executeQuery(projectContext, () => {
                            //Success
                        },
                            (sender, args) => {
                                throw new Error(`FAILED to load Content Type fields. Error: '${args.get_message()}'.`);
                            });


                        // Get the fields from Schema XML
                        //let fields = xmlDoc.getElementsByTagName("FieldRef");
                        let fields = xmlDoc.getElementsByTagName("Field");
                        let addedFields = false;
                        for (let i = 0; i < fields.length; i++) {
                            let fieldInternalName = fields[i].getAttribute('Name');

                            let ctColumn = this.selectFieldByName(existingContentTypeFields, fieldInternalName);
                            if (ctColumn == null) {
                                //Need to add field to content type
                                addedFields = true;
                                this.appendLog(`Adding field '${fieldInternalName}' to Content Type '${contentType.Name}'.`);
                                let existingColumn = this.selectFieldByName(siteColumns, fieldInternalName);

                                let fieldLinkCreationInfo = new SP.FieldLinkCreationInformation();

                                fieldLinkCreationInfo.set_field(existingColumn);

                                let fieldLink = existingContentType.get_fieldLinks().add(fieldLinkCreationInfo);
                                existingContentType.update(true);
                                projectContext.load(fieldLink);
                            }
                        }

                        if (addedFields) {
                            await this.executeQuery(projectContext, () => {
                                this.appendLog(`Updated fields on Content Type '${contentType.Name}'.`);
                            },
                                (sender, args) => {
                                    throw new Error(`FAILED to add fields to Content Type '${contentType.Name}'. Error: '${args.get_message()}'.`);
                                });
                        }
                        //else {
                        //No need for log, just makes it messy
                        //this.appendLog(`No updates required on Content Type '${contentType.Name}'.`);
                        //}
                    }
                }
                catch (error) {
                    this.appendError(error.toString()); //Log error and move on to next Context Type
                }
            }

            index++;
            progress = Math.round((index / count) * 100);
            this.updateProgress(progress);
        }

        //resolve();
        //}
        //catch (error) {
        //reject(error);
        //}
        //});
    }

    selectContentTypeByStringId(existingContentTypes, contentTypeStringId) {
        const contentTypeEnumerator = existingContentTypes.getEnumerator();
        while (contentTypeEnumerator.moveNext()) {
            let existing = contentTypeEnumerator.get_current();

            if (existing.get_stringId() == contentTypeStringId) {
                return existing;
            }
        }
        return null;
    }

    selectFieldByName(fields, internalName) {
        const siteColumnEnumerator = fields.getEnumerator();
        while (siteColumnEnumerator.moveNext()) {
            let existing = siteColumnEnumerator.get_current();

            if (existing.get_internalName() === internalName) {
                return existing;
            }
        }
        return null;
    }
}
