// Ensure the global SP and PS variables are defined
/// <reference path="PS.debug.js" />
/// <reference path="SP.debug.js" />
/// <reference path="SP.Runtime.js" />
var SP = SP || {};
var PS = PS || {};

class IOCustomFields extends IOBaseEntity {
    constructor(apiUrl, fileInput, logTextArea, progressBar) {
        super(apiUrl, fileInput, logTextArea, progressBar);
    }
    emptyDefaultValueId = "00000000-0000-0000-0000-000000000000";

    get_name() {
        return "Custom Fields";
    }

    customeFieldProperties() {
        return "Id,Name,Description,EntityType,EntityType,FieldType,DepartmentId,Formula,IsEditableInVisibility,IsMultilineText,IsRequired,IsWorkflowControlled,LookupAllowMultiSelect,LookupDefaultValue,IsLeafOnly,RollsDownToAssignments,RollupType,ShowToolTip,UseFieldForMatching,GraphicalIndicatorNonSummary,GraphicalIndicatorProjectSummary,GraphicalIndicatorSummary,EntityType";
    }

    async export() {
        this.startNewLog("Retrieving Custom Fields");
        //Using REST
        const data = await this.request('GET', `/_api/ProjectServer/CustomFields?$select=${this.customeFieldProperties()}&$expand=LookupTable,EntityType,Department`);
        //const customFields = data.value;
        const customFields = data.d.results;
        const lookupTableData = await this.request('GET', '/_api/ProjectServer/LookupTables?$select=Id,Name'); //for reference data
        //const lookupTables = lookupTableData.value;
        const lookupTables = lookupTableData.d.results;

        var count = customFields.length;

        //if a custom field has a lookup table and a default value, we need to extract that and include it in the exported data
        var index = 0;
        var progress = 0;
        for (const customField of customFields) {
            if (customField.hasOwnProperty('LookupTable')) {
                if (customField.LookupDefaultValue != this.emptyDefaultValueId) {
                    //Get the lookup Table by name
                    const lookupTable = this.getLookupByName(lookupTables, customField.LookupTable.Name);
                    if (lookupTable == null) {
                        throw new Error(`Could not locate Lookup Table '${customField.LookupTable.Name}' for Custom Field '${customField.Name}'`);
                    }

                    const lookupEntryData = await this.request('GET', `/_api/ProjectServer/LookupTables('${lookupTable.Id}')/Entries?$select=Id,FullValue`);
                    //const lookupEntries = lookupEntryData.value;
                    const lookupEntries = lookupEntryData.d.results;
                    var entry = this.getLookupEntryById(lookupEntries, customField.LookupDefaultValue);
                    if (entry == null) {
                        throw new Error(`Could not locate Default Value for Custom Field '${customField.Name}' in Lookup Table '${lookupTable.Name}'`);
                    }
                    customField.LookupDefaultFullValue = entry.FullValue;

                }
            }
            index++;
            progress = Math.round((index / count) * 100);
            this.updateProgress(progress);
        }

        this.appendLog(`Found ${count} Custom Fields in PWA`);

        var pwaConfig = {
            CustomFields: customFields
        };

        const blob = new Blob([JSON.stringify(pwaConfig, null, 4)], { type: "application/json" });
        const url = URL.createObjectURL(blob);
        const link = document.createElement('a');
        link.href = url;
        link.download = 'CustomFields.json';
        link.click();
    }

    getLookupEntryById(lookupTableEntries, id) {
        for (const entry of lookupTableEntries) {
            if (entry.Id == id) {
                return entry;
            }
        }
        return null;
    }

    getLookupEntryByFullValue(lookupTableEntries, fullValue) {
        for (const entry of lookupTableEntries) {
            if (entry.FullValue == fullValue) {
                return entry;
            }
        }
        return null;
    }

    getLookupByName(lookupTables, name) {
        for (const table of lookupTables) {
            if (table.Name === name) {
                //this.appendLog(`return lookup table with id='${table.Id}' and name='${table.Name}' for requested name='${name}'`);
                return table;
            }
        }
        return null;
    }

    async import() {
        //return new Promise(async (resolve, reject) => {
        //try {
        this.startNewLog('');
        const file = this.openFile();
        const fileContent = await this.readFileAsText(file);

        const pwaConfig = JSON.parse(fileContent);
        var customFields = pwaConfig.CustomFields;
        if (customFields == null) {
            reject(`No '${this.get_name()}' were found in the selected file.`);
        }

        var count = customFields.length;
        this.appendLog(`Found ${count} Custom Fields in file`);
        let projectContext = PS.ProjectContext.get_current();
        this.appendLog("Reading Custom Fields from PWA");
        var existingCustomFields = projectContext.get_customFields();
        var loadEntityTypes = projectContext.get_entityTypes();
        var projectEntity = loadEntityTypes.get_projectEntity();
        var taskEntity = loadEntityTypes.get_taskEntity();
        var resourceEntity = loadEntityTypes.get_resourceEntity();
        var assignmentEntity = loadEntityTypes.get_assignmentEntity();

        projectContext.load(existingCustomFields, `Include(${this.customeFieldProperties()})`);
        projectContext.load(loadEntityTypes);
        projectContext.load(projectEntity);
        projectContext.load(taskEntity);
        projectContext.load(resourceEntity);
        projectContext.load(assignmentEntity);

        await this.executeQuery(projectContext, async () => { /*success*/ }
            , (sender, args) => {//Failed
                //reject(`FAILED to read Custom Fields from PWA'. Error: '${args.get_message()}'.`);
                throw new Error(`FAILED to read Custom Fields from PWA'. Error: '${args.get_message()}'.`);
            }
        );

        var entityTypes = [projectEntity, taskEntity, resourceEntity, assignmentEntity];

        this.appendLog(`Found ${existingCustomFields.get_count()} Custom Fields in PWA`);

        var existingCustomField = null;

        var progress = 0;
        var fieldUpdated = false;
        var postponedImports = [];
        var customFieldHasAnUnImportedDependency = false;
        var index = 0;
        var count = customFields.length;
        //while (index < count) {
        while (customFields.length > 0) {
            postponedImports = [];
            for (const customField of customFields) {
                if (!(customField.Selected == null || customField.Selected == true)) {
                    this.appendLog(`Custom Field '${customField.Name}' is not selected for Import.`);
                }
                else {
                    customFieldHasAnUnImportedDependency = await this.customFieldHasDependenciesOnNonExistingFields(projectContext, customFields, existingCustomFields, customField);
                    if (!customFieldHasAnUnImportedDependency) {
                        fieldUpdated = false;
                        existingCustomField = await this.getMatchingExistingCustomField(projectContext, existingCustomFields, customField);

                        //try {
                        fieldUpdated = await this.importCustomField(projectContext, customField, entityTypes, existingCustomField, existingCustomFields, false);

                        //} catch (error) {
                        //    this.appendError(`Error occurred on Custom Field '${customField.Name}': ${error}`);
                        //}

                        if (fieldUpdated) {
                            //reload existing custom fields
                            existingCustomFields = projectContext.get_customFields();
                            projectContext.load(existingCustomFields, `Include(${this.customeFieldProperties()})`);
                            await this.executeQuery(projectContext, async () => { /*success*/ }
                                , (sender, args) => {//Failed
                                    //reject(`FAILED to read Custom Fields from PWA'. Error: '${args.get_message()}'.`);
                                    throw new Error(`FAILED to read Custom Fields from PWA'. Error: '${args.get_message()}'.`);
                                }
                            );
                        }

                        index++;
                        progress = Math.round((index / count) * 100);
                        this.updateProgress(progress);
                    }
                    else {
                        postponedImports.push(customField);
                        this.appendLog(`Postponing CustomField '${customField.Name}' as it has dependencies on`)
                    }
                }
            }//end customFields loop

            customFields = postponedImports;
            postponedImports = null;
        }

        //resolve();
        //    }
        //    catch
        //    {
        //        reject(error);
        //    }
        //});
    }

    async importCustomField(projectContext, customField, entityTypes, existingCustomField, existingCustomFields, useNewIdOnNewItems) {
        
        var fieldUpdated = false;
        try {

            if (existingCustomField == null) {
                this.appendLog(`Custom Field '${customField.Name}' does not exist, Creating it.`);
                fieldUpdated = true;
                var entityType = this.getEntityType(entityTypes, customField.EntityType.ID);
                var newCustomField = await this.createCustomField(projectContext, customField, entityType, useNewIdOnNewItems);
                existingCustomFields.add(newCustomField);
            }
            else {
                //Update existing custom Field
                fieldUpdated = await this.updateCustomField(projectContext, customField, existingCustomField);
            }

            if (fieldUpdated) {
                var addUpdateText = "update";
                if (existingCustomField == null) {
                    addUpdateText = "create";
                }

                existingCustomFields.update();
                projectContext.load(existingCustomFields);
                await this.executeQuery(projectContext, () => {
                    this.appendLog(`Custom Field '${customField.Name}' has been ${addUpdateText}d.`);
                },
                    (sender, args) => {
                        throw new Error(`FAILED to ${addUpdateText} Custom Field '${customField.Name}'. Error: '${args.get_message()}'.`);
                    }
                );
            }
            else {
                this.appendLog(`No change required for Custom Field '${customField.Name}'`);
            }
        }
        catch (error) {
            if (!useNewIdOnNewItems && error.toString().includes("PJClientCallableException: CustomFieldInvalidUID")) {
                //retry this methos using new Id
                this.appendWarning(`WARNING: Custom Field '${customField.Name} could not be created with its existing ID'. Retrying with a new ID.`)
                fieldUpdated = await this.importCustomField(projectContext, customField, entityTypes, existingCustomField, existingCustomFields, true);
            }
            else {
                this.appendError(`Error: '${error}'.${error.stack}`);
                fieldUpdated = false;
                //throw new Error(error.toString());
            }
        }
        return fieldUpdated;
    }

    async getMatchingExistingCustomField(projectContext, existingCustomFields, customField) {
        
        var existingCustomField = null;
        const fieldEnumerator = existingCustomFields.getEnumerator();
        var entityType = null;

        while (fieldEnumerator.moveNext()) {
            let existing = fieldEnumerator.get_current();
            entityType = existing.get_entityType();
            projectContext.load(entityType); //, 'Include(Id)');
            await this.executeQuery(projectContext, async () => { /*success*/ },
                (sender, args) => {
                    throw new Error(`FAILED to load EntityType for Custom Field '${customField.Name}'. Error: '${args.get_message()}'.`);
                }
            );
            if ((existing.get_name() == customField.Name) && (entityType.get_id() == customField.EntityType.ID)) {
                
                return existing;
                break;
            }
        }
        
        return null;
        
    }

    getEntityType(entityTypes, id) {
        for (const entityType of entityTypes) {
            //this.appendLog(`entityType.get_id() = '${entityType.get_id()}', id='${id}'`);
            if (entityType.get_id() == id) { //do not use ===, because the types do not match
                //this.appendLog(`returning entityType.get_id() = '${entityType.get_id()}'`);
                return entityType;
            }
        }
        return null;
    }

    async customFieldExists(projectContext, existingCustomFields, customField) {
        //return new Promise(async (resolve, reject) => {
        //const fieldEnumerator = existingCustomFields.getEnumerator();
        //console.log(`Does '${customField.Name}' exist already?`);
        var existing = await this.getMatchingExistingCustomField(projectContext, existingCustomFields, customField);
        let answer = (existing != null);
        console.log(`Does '${customField.Name}' exist already? A=${answer}`);
        return answer;
        //});
    }

    async customFieldHasDependenciesOnNonExistingFields(projectContext, customFields, existingCustomFields, customFieldToImport) {
        var exists = false;

        if (customFieldToImport.Formula) {
            if (customFieldToImport.Formula != null) {
                console.log(`Custom Field '${customFieldToImport.Name}' has a formula, checking if the formula contains any custom field that does not yet exist.`);
                for (const customField of customFields) {
                    if (customFieldToImport.EntityType.Name == customField.EntityType.Name) {
                        if (customFieldToImport.Formula.includes(`[${customField.Name}]`)) { //TODO: Might need to include a filter on EntityType here too

                            exists = await this.customFieldExists(projectContext, existingCustomFields, customField);
                            if (!exists) {
                                //The customFieldToImport contains a Formula dependency on another custom field that has not yet been imported
                                console.log(`Custom Field '${customFieldToImport.Name}' depends on '${customField.Name}' which does not exist yet.`);
                                return true;
                            }
                            else {
                                console.log(`Custom Field '${customFieldToImport.Name}' depends on '${customField.Name}' but thats ok, it exists already.`);
                            }

                        } //else does not depend on this field anyway
                        else {
                            console.log(`Custom Field '${customFieldToImport.Name}' does NOT depends on '${customField.Name}'.`);
                        }

                    } //else not the same entity type
                    else {
                        console.log(`Custom Field '${customFieldToImport.Name}[${customFieldToImport.EntityType.Name}]' is not the same EntityType as '${customField.Name}[${customField.EntityType.Name}]'.`);
                    }
                }
            }
        }
        console.log(`customFieldHasDependenciesOnNonExistingFields() is returning false for Custom Field '${customFieldToImport.Name}'.`);
        return false;
    }

    async createCustomField(projectContext, customField, entityType, useNewId) {
        //return new Promise(async (resolve, reject) => {
        //var newGuid = SP.Guid.newGuid().toString();
        var newGuid = customField.Id;
        if (useNewId) {
            newGuid = SP.Guid.newGuid().toString();
            this.appendLog(`Defined new Id '${newGuid}' for custom field '${customField.Name}'`);
        }
        //this.appendLog('New Id is :' + newGuid);
        //try {
        var newCustomField = new PS.CustomFieldCreationInformation();
        newCustomField.set_id(newGuid);

        newCustomField.set_name(customField.Name);
        newCustomField.set_description(customField.Description);

        newCustomField.set_fieldType(customField.FieldType);
        //newCustomField.set_entityType(customField.EntityType); //.ID);
        //this.appendWarning(`Setting Entity Type ID='${entityType.get_id()}'`);

        newCustomField.set_entityType(entityType);
        newCustomField.set_formula(customField.Formula);
        newCustomField.set_isEditableInVisibility(customField.IsEditableInVisibility);
        newCustomField.set_isMultilineText(customField.IsMultilineText);
        newCustomField.set_isRequired(customField.IsRequired);
        newCustomField.set_isWorkflowControlled(customField.IsWorkflowControlled);
        newCustomField.set_lookupAllowMultiSelect(customField.LookupAllowMultiSelect);
        newCustomField.set_isLeafOnly(customField.IsLeafOnly);

        if (customField.GraphicalIndicatorNonSummary != null) {
            newCustomField.set_graphicalIndicatorNonSummary(customField.GraphicalIndicatorNonSummary);
        }
        if (customField.GraphicalIndicatorProjectSummary != null) {
            newCustomField.set_graphicalIndicatorProjectSummary(customField.GraphicalIndicatorProjectSummary);
        }
        if (customField.GraphicalIndicatorSummary != null) {
            newCustomField.set_graphicalIndicatorSummary(customField.GraphicalIndicatorSummary);
        }

        if (customField.hasOwnProperty('LookupTable') && customField.LookupTable.hasOwnProperty('Name')) {
            //get LookupTable object from PWA by JSOM
            const existingLookupTable = await this.getExistingLookupTableForCustomField(projectContext, customField);
            newCustomField.set_lookupTable(existingLookupTable);

            if (customField.LookupDefaultValue != this.emptyDefaultValueId) {
                const lookupEntryData = await this.request('GET', `/_api/ProjectServer/LookupTables('${existingLookupTable.get_id()}')/Entries?$select=Id,FullValue`);
                var entry = this.getLookupEntryByFullValue(lookupEntryData.d.results, customField.LookupDefaultFullValue);
                if (entry == null) {
                    throw new Error(`Could not locate Default Value '{$customField.LookupDefaultFullValue}' for Custom Field '${customField.Name}' in Lookup Table '${existingLookupTable.get_name()}'`);
                }
                newCustomField.set_lookupDefaultValue(entry.Id);
            }
        }
        //resolve(newCustomField);
        return newCustomField

        //}
        //catch (error) {
        //    reject(error);
        //}
        //});
    }

    async updateCustomField(projectContext, customField, existingCustomField) {
        return new Promise(async (resolve, reject) => {
            //try {
            var updated = false;
            var propertiesUpdated = '';
            //existingCustomField.set_id(customField.Id); //Must use a new Id :(
            //existingCustomField.set_name(customField.Name);
            if (existingCustomField.get_description() != customField.Description) {
                existingCustomField.set_description(customField.Description);
                propertiesUpdated = propertiesUpdated + ',Description';
                updated = true;
            }

            //These cannot be changed
            if (existingCustomField.get_fieldType() != customField.FieldType) {
                //    existingCustomField.set_fieldType(customField.FieldType);
                //    updated = true;
                this.appendError(`Cannot update FieldType on Custom Field '${customField.Name}'. Property is read-only.`);
            }
            if (existingCustomField.get_entityType().get_id() != customField.EntityType.ID) {
                //    existingCustomField.set_entityType(customField.EntityType.ID);
                //    updated = true;
                //Should get to this code anyway.
                this.appendError(`Cannot update EntityType on Custom Field '${customField.Name}'. Property is read-only.`);
            }

            if (existingCustomField.get_formula() != customField.Formula) {
                existingCustomField.set_formula(customField.Formula);
                propertiesUpdated = propertiesUpdated + ',Formula';
                updated = true;
            }
            if (existingCustomField.get_isEditableInVisibility() != customField.IsEditableInVisibility) {
                existingCustomField.set_isEditableInVisibility(customField.IsEditableInVisibility);
                propertiesUpdated = propertiesUpdated + ',IsEditableInVisibility';
                updated = true;
            }
            if (existingCustomField.get_isMultilineText() != customField.IsMultilineText) {
                existingCustomField.set_isMultilineText(customField.IsMultilineText);
                propertiesUpdated = propertiesUpdated + ',IsMultilineText';
                updated = true;
            }
            if (existingCustomField.get_isRequired() != customField.IsRequired) {
                existingCustomField.set_isRequired(customField.IsRequired);
                propertiesUpdated = propertiesUpdated + ',IsRequired';
                updated = true;
            }
            if (existingCustomField.get_isWorkflowControlled() != customField.IsWorkflowControlled) {
                existingCustomField.set_isWorkflowControlled(customField.IsWorkflowControlled);
                propertiesUpdated = propertiesUpdated + ',IsWorkflowControlled';
                updated = true;
            }
            if (existingCustomField.get_lookupAllowMultiSelect() != customField.LookupAllowMultiSelect) {
                existingCustomField.set_lookupAllowMultiSelect(customField.LookupAllowMultiSelect);
                propertiesUpdated = propertiesUpdated + ',LookupAllowMultiSelect';
                updated = true;
            }

            if (existingCustomField.get_isLeafOnly() != customField.IsLeafOnly) {
                existingCustomField.set_isLeafOnly(customField.IsLeafOnly);
                propertiesUpdated = propertiesUpdated + ',IsLeafOnly';
                updated = true;
            }
            if (existingCustomField.get_graphicalIndicatorNonSummary() != customField.GraphicalIndicatorNonSummary) {
                existingCustomField.set_graphicalIndicatorNonSummary(customField.GraphicalIndicatorNonSummary);
                propertiesUpdated = propertiesUpdated + ',GraphicalIndicatorNonSummary';
                updated = true;
            }
            if (existingCustomField.get_graphicalIndicatorProjectSummary() != customField.GraphicalIndicatorProjectSummary) {
                existingCustomField.set_graphicalIndicatorProjectSummary(customField.GraphicalIndicatorProjectSummary);
                propertiesUpdated = propertiesUpdated + ',GraphicalIndicatorProjectSummary';
                updated = true;
            }
            if (existingCustomField.get_graphicalIndicatorSummary() != customField.GraphicalIndicatorSummary) {
                existingCustomField.set_graphicalIndicatorSummary(customField.GraphicalIndicatorSummary);
                propertiesUpdated = propertiesUpdated + ',GraphicalIndicatorSummary';
                updated = true;
            }

            if (customField.hasOwnProperty('LookupTable') && customField.LookupTable.hasOwnProperty('Name')) {
                //get LookupTable object from PWA by JSOM
                const shouldUseLookupTable = await this.getExistingLookupTableForCustomField(projectContext, customField);

                if (existingCustomField.get_lookupTable() == null) {
                    //existingCustomField.set_lookupTable(shouldUseLookupTable);
                    //propertiesUpdated = propertiesUpdated + ',LookupTable';
                    //updated = true;
                    this.appendError(`Cannot update LookupTable on Custom Field '${customField.Name}'. Property is read-only.`);

                }
                else {
                    var existingCustomFieldLookupTable = existingCustomField.get_lookupTable();
                    projectContext.load(existingCustomFieldLookupTable);//, 'Include(Id,Name)');
                    await this.executeQuery(projectContext, () => { /*success*/ }, (sender, args) => { reject(`FAILED to load Lookup Table by id'${existingCustomFieldLookupTable.get_id()}'. Error: ${args.get_message()}`); });

                    if (existingCustomFieldLookupTable.get_name() != shouldUseLookupTable.get_name()) {
                        //existingCustomField.set_lookupTable(shouldUseLookupTable);
                        //propertiesUpdated = propertiesUpdated + ',LookupTable';
                        //updated = true;
                        this.appendError(`Cannot update LookupTable on Custom Field '${customField.Name}'. Property is read-only.`);
                    }
                }

                if (customField.LookupDefaultValue === this.emptyDefaultValueId) {
                    if (existingCustomField.get_lookupDefaultValue() != this.emptyDefaultValueId) {
                        existingCustomField.set_lookupDefaultValue(this.emptyDefaultValueId);
                        propertiesUpdated = propertiesUpdated + ',LookupDefaultValue';
                        updated = true;
                    }
                }
                else {
                    const lookupEntryData = await this.request('GET', `/_api/ProjectServer/LookupTables('${shouldUseLookupTable.get_id()}')/Entries?$select=Id,FullValue`);
                    var entry = this.getLookupEntryByFullValue(lookupEntryData.d.results, customField.LookupDefaultFullValue);
                    if (entry == null) {
                        throw new Error(`Could not locate Default Value '${customField.LookupDefaultFullValue}' for Custom Field '${customField.Name}' in Lookup Table '${shouldUseLookupTable.get_name()}'`);
                    }
                    if (existingCustomField.get_lookupDefaultValue() != entry.Id) {
                        existingCustomField.set_lookupDefaultValue(entry.Id);
                        propertiesUpdated = propertiesUpdated + ',LookupDefaultValue';
                        updated = true;
                    }
                }
            }
            else {
                //Lookup should be empty
                //Using JSOM to return what is supposed to be null for the existing custom field, doesn;t work. It returns an uninitialized object of some kind
                //var existingCustomFieldLookupTable = existingCustomField.get_lookupTable();
                //projectContext.load(existingCustomFieldLookupTable); //, 'Include(Id,Name)');
                //await this.executeQuery(projectContext, () => { /*success*/ }, (sender, args) => { reject(`FAILED to load Lookup Table by id'${existingCustomFieldLookupTable.get_id()}'. Error: ${args.get_message()}`); });
                //if (existingCustomFieldLookupTable != null) {

                //So we'll use the REST method instead
                const existingCustomFieldData = await this.request('GET', `/_api/ProjectServer/CustomFields('${existingCustomField.get_id()}')?$select=Id,Name&$expand=LookupTable`);

                if (existingCustomFieldData.hasOwnProperty('LookupTable')) {
                    //existingCustomField.set_lookupTable(null);
                    //propertiesUpdated = propertiesUpdated + ',LookupTable';
                    //updated = true;
                    this.appendError(`Cannot update LookupTable on Custom Field '${customField.Name}'. Property is read-only.`);
                }
            }
            if (updated) {
                this.appendLog(`On Custom Field '${customField.Name}' the following properties are being updated '${propertiesUpdated.substring(1)}'`);
            }
            resolve(updated);
            //return updated;

            //}
            //catch (error) {
            //    reject(error);
            //}
        });
    }

    async getExistingLookupTableForCustomField(projectContext, customField) {
        return new Promise(async (resolve, reject) => {
            //try {
            const lookupTableName = customField.LookupTable.Name;
            var existingTables = projectContext.get_lookupTables();
            projectContext.load(existingTables, 'Include(Name, Id)');

            await this.executeQuery(projectContext, () => {

                var existingTable = null;
                const tableEnumerator = existingTables.getEnumerator();
                while (tableEnumerator.moveNext()) {
                    let existing = tableEnumerator.get_current();

                    if (existing.get_name() == lookupTableName) {
                        resolve(existing);
                        return existing;
                    }
                }

                throw new Error(`FAILED to find Lookup Table '${lookupTableName}' for Custom Field '${customField.Name}'`);
                return null;
            },
                (sender, args) => {
                    throw new Error(`FAILED to load Lookup Tables from PWA. Error: '${args.get_message()}'.`);
                }
            );
            //}
            //catch (error) {
            //    reject(error);
            //}
        });

    }
}