// Ensure the global SP and PS variables are defined
/// <reference path="PS.debug.js" />
/// <reference path="SP.debug.js" />
/// <reference path="SP.Runtime.js" />
var SP = SP || {};
var PS = PS || {};

class IOLookupEntries extends IOBaseEntity {
    constructor(apiUrl, fileInput, logTextArea, progressBar) {
        super(apiUrl, fileInput, logTextArea, progressBar);
    }

    get_name() {
        return "Lookup Entries";
    }

    async export() {
        this.startNewLog("Retrieving Lookup Entries");
        //Using REST
        const data = await this.request('GET', '/_api/ProjectServer/LookupTables?$select=Id,Name,Masks,FieldType');
        //const lookupTables = data.value;
        const lookupTables = data.d.results;

        var count = lookupTables.length;
        this.appendLog(`Found ${count} Lookup Tables in PWA`);
        var index = 0;
        var progress = 0;
        var countOfEntries = 0;

        for (const table of lookupTables) {
            const data = await this.request('GET', `/_api/ProjectServer/LookupTables('${table.Id}')/Entries?$select=*`);
            //table.Entries = data.value;
            table.Entries = data.d.results;
            this.appendLog(`Exported ${table.Entries.length} Entries from the Lookup Table '${table.Name}' in PWA`);
            countOfEntries += table.Entries.length;
            index++;
            progress = Math.round((index / count) * 100);
            this.updateProgress(progress);
        }

        this.appendLog(`Exported ${countOfEntries} Entries across the ${count} Lookup Tables in PWA`);

        var pwaConfig = {
            LookupTableEntries: lookupTables
        };

        const blob = new Blob([JSON.stringify(pwaConfig, null, 4)], { type: "application/json" });

        const url = URL.createObjectURL(blob);
        const link = document.createElement('a');
        link.href = url;
        link.download = 'LookupEntries.json';
        link.click();
    }

    async import() {
        return new Promise(async (resolve, reject) => {
            try {
                this.startNewLog('');

                if (!confirm('** WARNING **\nLookup Table entries are completely deleted from the target environment before importing from file.\nExisting Entries will be merged in with the entries from file, before being deleted!\n\nAre you certain you wish to proceed?')) {
                    throw new Error('Import aborted');
                }

                this.appendLog("Reading Lookup Entries from file");
                
                const file = this.openFile();
                const fileContent = await this.readFileAsText(file);
                const pwaConfig = JSON.parse(fileContent);
                const lookupTables = pwaConfig.LookupTableEntries;
                if (lookupTables == null) {
                    reject(`No '${this.get_name()}' were found in the selected file.`);
                }

                var count = lookupTables.length;
                this.appendLog(`Found ${count} Lookup Tables in file`);
                let projectContext = PS.ProjectContext.get_current();
                this.appendLog("Reading Lookup Tables from PWA");
                var existingTables = projectContext.get_lookupTables();
                projectContext.load(existingTables, 'Include(Name, Id, Masks, FieldType)');

                await this.executeQuery(projectContext, () => {
                    this.appendLog(`Found ${existingTables.get_count()} Lookup Tables in PWA`);
                },
                    (sender, args) => {
                        reject(`FAILED to read Lookup Tables from PWA'. Error: '${args.get_message()}'.`);
                        //throw new Error(args.get_message());
                    }
                );

                var existingTable = null;
                var index = 0;
                var progress = 0;

                for (const table of lookupTables) {

                    if (!(table.Selected == null || table.Selected == true)) {
                        this.appendLog(`Entries for Lookup Table '${table.Name}' are not selected for Import.`);
                    }
                    else {
                        existingTable = null;
                        const tableEnumerator = existingTables.getEnumerator();
                        while (tableEnumerator.moveNext()) {
                            let existing = tableEnumerator.get_current();

                            if (existing.get_name() == table.Name) {
                                existingTable = existing;

                                break;
                            }
                        }

                        if (existingTable == null) {
                            this.appendWarning(`Lookup Table '${table.Name}' does not exist. Skipping.`);
                        } else {
                            await this.importLookupEntriesForTable(projectContext, table, existingTable, existingTables);
                        }
                    }

                    index++;
                    progress = Math.round((index / count) * 100);
                    this.updateProgress(progress);
                }//end of Tables loop
                resolve();
            }
            catch (error) {
                reject(error);
            }
        });
    }

    async importLookupEntriesForTable(projectContext, table, existingTable, existingTables) {
        return new Promise(async (resolve, reject) => {
            var builtInTablesToEdit = [];
            builtInTablesToEdit.push("Project Impact");
            builtInTablesToEdit.push("Relative Importance");

            if (builtInTablesToEdit.includes(table.Name)) {
                this.appendWarning(`Skipping Lookup Table '${table.Name}' because it is a BuiltIn table`);
                resolve();
                return;
            }

            //try {
            //Validate Table Masks match
            if (!this.doTableDefinitionsMatch(table, existingTable)) {
                resolve();
                return;
            }

            const data = await this.request('GET', `/_api/ProjectServer/LookupTables('${existingTable.get_id()}')/Entries?$select=*`);
            //var existingEntriesByRest = data.value;
            var existingEntriesByRest = data.d.results;

            //Merge existing and to be imported
            var entriesToAdd = this.mergeEntries(table.Entries, existingEntriesByRest);

            //this.appendLog(`entriesToAdd = ${JSON.stringify(entriesToAdd)}`);

            await this.deleteLookupEntriesForTable(projectContext, existingTable, existingTables);

            var madeTableUpdates = false;
            const masks = table.Masks.results;

            var existingEntries = await this.reloadLookupEntriesForTable(projectContext, existingTable);
            //existingEntries should be empty at this point, because we deleted all entries earlier

            //var entriesToAdd = this.getEntriesToAdd(table.Entries, existingEntries);

            //Add missing Entries (new items)
            for (const entry of entriesToAdd) {

                //Entry does not exist, so lets create it
                //this.appendLog(`Lookup Entry '${entry.FullValue}' does not exist, Creating it.`);

                var ltEntry = new PS.LookupEntryCreationInformation();

                ltEntry.set_id(entry.Id);
                ltEntry.set_description(entry.Description);
                ltEntry.set_sortIndex(entry.SortIndex);

                if (masks.length == 1) {
                    ltEntry.set_parentId(null);
                }
                else {
                    var parentId = this.getLookupEntryParentIdFromFile(table.Entries, entry.FullValue, masks);
                    if (parentId == null) {
                        ltEntry.set_parentId(null);
                    }
                    else {
                        //Get actual Parent id from PWA, it may be different
                        parentId = this.getLookupEntryParentIdFromPWA(existingEntries, entry.FullValue, masks);
                        ltEntry.set_parentId(parentId);
                    }
                }

                var ltValue = new PS.LookupEntryValue();

                //FieldType is derived from the Tables field Type
                switch (table.FieldType) {
                    case (PS.CustomFieldType.TEXT):
                        ltValue.set_textValue(entry.Value);
                        break;
                    case (PS.CustomFieldType.COST):
                        ltValue.set_numberValue(entry.Value);
                        break;
                    case (PS.CustomFieldType.DURATION):
                        ltValue.set_durationValue(entry.Value);
                        break;
                    case (PS.CustomFieldType.DATE):
                        ltValue.set_dateValue(entry.Value);
                        break;
                    case (PS.CustomFieldType.NUMBER):
                        ltValue.set_numberValue(entry.Value);
                        break;
                }

                ltEntry.set_value(ltValue);

                existingEntries.add(ltEntry);

                existingTable.entries = existingEntries;

                await this.saveLookupEntriesForTable(projectContext, existingTables, `Added/updated Lookup Entry ${entry.FullValue} for '${table.Name}' has been added.`, `FAILED to add new Lookup Entry '${entry.FullValue}' for Lookup Table '${table.Name}'`);
                //madeTableUpdates = true;

                //Reload LookupEntries after each save
                existingEntries = await this.reloadLookupEntriesForTable(projectContext, existingTable);

            }//End of Entries to Add loop

            //var entriesToUpdate = this.getEntriesToUpdate(table.Entries, existingEntries);
            ////Update all (new and pre-existing) Entries, this "should" fix any sort order problems
            //for (const entry of table.Entries) {

            //    var existingEntry = null;
            //    const entryEnumerator = existingEntries.getEnumerator();
            //    while (entryEnumerator.moveNext()) {
            //        let existing = entryEnumerator.get_current();

            //        if (existing.get_id() == entry.Id) { //should have no issue here after we re-loaded after the add step above
            //            existingEntry = existing;

            //            break;
            //        }
            //    }

            //    if (existingEntry != null) {
            //        //Entry does exist, so we update it

            //        if (existingEntry.get_description() != entry.Description) {

            //            existingEntry.set_description(entry.Description);
            //            this.appendLog(`Lookup Entry '${entry.FullValue}' has been updated.`);

            //            madeTableUpdates = true;
            //        }
            //    }
            //}//end of entries to update loop

            //await this.saveLookupEntriesForTable(projectContext, existingTables, `Lookup Entries for Lookup Table '${table.Name}' saved.`, `FAILED to save Lookup Entries for Lookup Table '${table.Name}'`);

            resolve();
            //}
            //catch (error) {
            //    reject(error);
            //}
        });
    }

    mergeEntries(newEntries, existingEntries) {
        var entryToAdd = null;
        for (const existingEntry of existingEntries) {
            entryToAdd = this.getEntryByFullValue(newEntries, existingEntry.FullValue);
            if (entryToAdd == null) {
                //Add the existing entry to the newEntries to import
                //this.appendLog(`Adding existing entry ${existingEntry.FullValue} to entries to import`)
                newEntries.push(existingEntry);
            }
            else {
                entryToAdd.Id = existingEntry.Id; //ensure the same entry has the same Id
            }
        }

        //this.appendLog(`Sorting ${newEntries.length} Lookup Entries`);
        newEntries = newEntries.sort(this.compareEntriesByFullValue);

        //Re-calculate SortIndex on entries
        //this.appendLog(`Re-calculating SortIndex for ${newEntries.length} Lookup Entries`);
        var sortIndex = 1;
        for (const entry of newEntries) {
            entry.SortIndex = sortIndex.toFixed(10);
            sortIndex++;
        }

        return newEntries;
    }

    //reapplyEntrySortIndexes2(projectContext, existingEntries, existingTables, existingTable) {

    //    return new Promise(async (resolve, reject) => {
    //        //try {
    //        //reverse order
    //        var count = existingEntries.get_count();

    //        var newSortIndex = 0;

    //        var item = null;
    //        while (count > 0) {
    //            newSortIndex = count.toFixed(10);
    //            item = existingEntries.getItemAtIndex(count - 1);
    //            var previousSortIndex = item.get_sortIndex();
    //            if (previousSortIndex != newSortIndex) {
    //                item.set_sortIndex(newSortIndex);

    //                this.appendLog(`Updating SortIndexes for Lookup Table '${existingTable.get_name()}'. Setting ${previousSortIndex} to ${newSortIndex}`);

    //                //may have to do a save for each item change
    //                await this.saveLookupEntriesForTable(projectContext, existingTables, `Updated SortIndexes for Lookup Table '${existingTable.get_name()}'. Setting ${previousSortIndex} to ${newSortIndex}`, `FAILED to update SortIndex for Lookup Table '${existingTable.get_name()}'`);
    //                existingEntries = await this.reloadLookupEntriesForTable(projectContext, existingTable);
    //            }

    //            count--;
    //        }

    //        resolve();
    //        //}
    //        //catch (error) {
    //        //    reject(error);
    //        //}
    //    });
    //}

    //reapplyEntrySortIndexes(projectContext, existingEntries, existingTables, existingTable) {

    //    return new Promise(async (resolve, reject) => {
    //        //try {
    //        //reverse order
    //        var count = existingEntries.get_count();
    //        var newSortIndex = 0;
    //        var index = 0;
    //        var item = null;
    //        while (index < count) {
    //            item = existingEntries.getItemAtIndex(index);
    //            newSortIndex = (index + 1).toFixed(10);

    //            var previousSortIndex = item.get_sortIndex();
    //            if (parseFloat(previousSortIndex) != parseFloat(newSortIndex)) {
    //                item.set_sortIndex(newSortIndex);

    //                this.appendLog(`Updating SortIndexes for Lookup Table '${existingTable.get_name()}'. Setting ${previousSortIndex} to ${newSortIndex}`);

    //                //may have to do a save for each item change
    //                await this.saveLookupEntriesForTable(projectContext, existingTables, `Updated SortIndexes for Lookup Table '${existingTable.get_name()}'. Setting ${previousSortIndex} to ${newSortIndex}`, `FAILED to update SortIndex for Lookup Table '${existingTable.get_name()}'`);
    //                existingEntries = await this.reloadLookupEntriesForTable(projectContext, existingTable);
    //            }
    //            else {
    //                console.log(`NO Update: previousIndex = ${previousSortIndex}, newIndex = ${newSortIndex}`);
    //            }

    //            index++;
    //        }

    //        resolve();
    //        //}
    //        //catch (error) {
    //        //    reject(error);
    //        //}
    //    });
    //}

    //reapplyEntrySortIndexes(pwaEntries) {
    //forwards
    //    var sortIndex = 1;
    //    const entryEnumerator = pwaEntries.getEnumerator();
    //    while (entryEnumerator.moveNext()) {
    //        let existing = entryEnumerator.get_current();
    //        existing.set_sortIndex(sortIndex);
    //        sortIndex++;
    //    }
    //}

    async deleteLookupEntriesForTable(projectContext, existingTable, existingTables) {
        return new Promise(async (resolve, reject) => {
            var existingEntries = await this.reloadLookupEntriesForTable(projectContext, existingTable);
            var existingEntry = null;
            var entriesToDelete = [];
            const entryEnumerator = existingEntries.getEnumerator();
            while (entryEnumerator.moveNext()) {
                let existing = entryEnumerator.get_current();
                entriesToDelete.push(existing);
            }
            for (let i = 0; i < entriesToDelete.length; i++) {
                entriesToDelete[i].deleteObject();
            }

            await this.saveLookupEntriesForTable(projectContext, existingTables, `Deleted all lookup entries for table '${existingTable.get_name()}'`, `FAILED to delete lookup entries for table '${existingTable.get_name()}'`);

            resolve();
        });
    }


    async reloadLookupEntriesForTable(projectContext, existingTable) {
        return new Promise(async (resolve, reject) => {
            //try {
            var existingEntries = existingTable.get_entries();
            projectContext.load(existingEntries, 'Include(Id, FullValue,SortIndex, Value, Description)');
            await this.executeQuery(projectContext, () => {
                resolve(existingEntries);
            },
                (sender, args) => {
                    reject(`FAILED to load Lookup Entries for Lookup Table '${table.Name}'. Error: '${args.get_message()}'.`);
                }
            );
            //    }
            //    catch (error) {
            //        reject(error);
            //    }
        });
    }

    async saveLookupEntriesForTable(projectContext, existingTables, successMessage, failMessage) {
        return new Promise(async (resolve, reject) => {
            //try {
            existingTables.update();
            await this.executeQuery(projectContext, () => {
                this.appendLog(successMessage);
                resolve();
            },
                (sender, args) => {
                    reject(`${failMessage}. Error: '${args.get_message()}'.`);
                }
            );
            //}
            //catch (error) {
            //    reject(error);
            //}
        });
    }

    doTableDefinitionsMatch(table, existingTable) {

        if (table.FieldType != existingTable.get_fieldType()) {
            this.appendWarning(`Skipping Lookup Table '${table.Name}' because the FieldType does not match between the file version and the PWA version`);
            return false;
        }

        const masks = table.Masks.results;
        const existingMasks = existingTable.get_masks();

        if (masks.length != existingMasks.length) {
            this.appendWarning(`Skipping Lookup Table '${table.Name}' because the Masks do not match between the file version and the PWA version`);
            return false;
        }

        var mask = null;
        var existingMask = null;
        for (let i = 0; i < masks.length; i++) {
            mask = masks[i];
            existingMask = existingMasks[i];

            if (mask.Separator != existingMask.get_separator()) {
                this.appendWarning(`Skipping Lookup Table '${table.Name}' because the Masks do not match between the file version and the PWA version`);
                return false;
            }
            if (mask.Length != existingMask.get_length()) {
                this.appendWarning(`Skipping Lookup Table '${table.Name}' because the Masks do not match between the file version and the PWA version`);
                return false;
            }
            if (mask.MaskType != existingMask.get_maskType()) {
                this.appendWarning(`Skipping Lookup Table '${table.Name}' because the Masks do not match between the file version and the PWA version`);
                return false;
            }
        }

        //If not, return false
        return true;
    }

    //getFullValueParts(fullValue, masks) {
    //    if (fullValue === '' || fullValue === null) {
    //        throw new Error(`fullValue cannot be null or an empty string`);
    //    }
    //    var parts = [];
    //    var positionIndex = 0;
    //    //Create parts
    //    for (let i = 0; i < masks.length; i++) {
    //        var mask = masks[i];
    //        var separator = mask.Separator;
    //        positionIndex = fullValue.indexOf(separator);
    //        if (positionIndex > -1) {
    //            var partValue = fullValue.substring(0, positionIndex);
    //            parts.push(partValue);
    //            fullValue = fullValue.substring(positionIndex + separator.length);
    //        }
    //        else {
    //            parts.push(fullValue); //last part
    //            break;
    //        }
    //    }
    //    return parts;
    //}

    //This method returns the fullValue of an entries parent calculated by the entries fullValue
    //It expects the masks parameter to be provided from a file source (i.e. REST response)
    getLookupEntryParentFullValue(fullValue, masks) {
        if (masks.length == 1) {
            //This is a single level. There are no parents
            return null;
        }
        if (fullValue === '' || fullValue === null) {
            throw new Error(`fullValue cannot be null or an empty string`);
        }
        var originalValue = fullValue;
        //var maskLevel = 0;
        var parts = [];
        var separators = [];
        var positionIndex = 0;
        //Create parts
        for (let i = 0; i < masks.length; i++) {
            var mask = masks[i];
            var separator = mask.Separator;
            positionIndex = fullValue.indexOf(separator);
            if (positionIndex > -1) {
                var partValue = fullValue.substring(0, positionIndex);
                parts.push(partValue);
                separators.push(separator);
                fullValue = fullValue.substring(positionIndex + separator.length);
            }
            else {
                //console.log(`Mask '${separator}' does not exist in string`);
                break;
            }
        }
        //Build parent fullValue from parts
        var parentFullValue = '';
        var count = parts.length;
        if (count == 0) {
            return null; //parent is root
        }
        var separatorPrefix = '';
        for (let i = 0; i < count; i++) {
            parentFullValue = parentFullValue.concat(separatorPrefix, parts[i]);
            separatorPrefix = separators[i];
        }

        return parentFullValue;
    }

    //This method returns the parentId of an entry based on its fullValue
    //It expects the entries and masks parameters to be provided from a file source (i.e. REST response)
    getLookupEntryParentIdFromFile(entries, fullValue, masks) {
        if (masks.length == 1) {
            //This is a single level. There are no parents
            return null;
        }

        var parentFullValue = this.getLookupEntryParentFullValue(fullValue, masks);

        if (parentFullValue == null) {
            //There is no parent
            return null;
        }

        for (const entry of entries) {
            if (entry.FullValue == parentFullValue) {
                return entry.Id;
            }
        }
        throw new Error(`Could not find parent Lookup Entry for ${fullValue}, parentFullValue=${parentFullValue}`);
    }

    //This method returns the parentId of an entry based on its fullValue
    //It expects the entries and masks parameters to be provided from a PWA source (i.e. JSOM response)
    getLookupEntryParentIdFromPWA(pwaEntries, fullValue, masks) {
        if (masks.length == 1) {
            //This is a single level. There are no parents
            return null;
        }

        var parentFullValue = this.getLookupEntryParentFullValue(fullValue, masks);

        if (parentFullValue == null) {
            //There is no parent
            return null;
        }

        const entryEnumerator = pwaEntries.getEnumerator();
        while (entryEnumerator.moveNext()) {
            let existing = entryEnumerator.get_current();
            if (existing.get_fullValue() == parentFullValue) {
                return existing.get_id();
            }
        }
        throw new Error(`Could not find parent Lookup Entry for ${fullValue}, parentFullValue=${parentFullValue}`);
    }

    //getEntriesToUpdate(fileEntries, pwaEntries) {
    //    const entriesToUpdate = [];
    //    const entryEnumerator = pwaEntries.getEnumerator();
    //    while (entryEnumerator.moveNext()) {
    //        let existing = entryEnumerator.get_current();

    //        for (const entry of fileEntries) {
    //            if (existing.get_fullValue() === entry.FullValue) {
    //                entriesToUpdate.push(entry);
    //                break;
    //            }
    //        }
    //    }
    //    return entriesToUpdate;
    //}

    //getEntriesToAdd(fileEntries, pwaEntries) {
    //    const entriesToAdd = [];

    //    var entryToAdd = null;
    //    var entryExists = false;
    //    for (const entry of fileEntries) {
    //        entryExists = false;
    //        const entryEnumerator = pwaEntries.getEnumerator();
    //        while (entryEnumerator.moveNext()) {
    //            let existing = entryEnumerator.get_current();

    //            if (existing.get_fullValue() === entry.FullValue) {
    //                entryExists = true;
    //                break;
    //            }
    //        }
    //        if (!entryExists) {
    //            entriesToAdd.push(entry);
    //        }
    //    }

    //    entriesToAdd.sort(this.compareEntriesBySortIndex);

    //    return entriesToAdd;
    //}

    //compareEntriesBySortIndex(a, b) {
    //    if (parseFloat(a.SortIndex) < parseFloat(b.SortIndex)) {
    //        return -1;
    //    }
    //    if (parseFloat(a.SortIndex) > parseFloat(b.SortIndex)) {
    //        return 1;
    //    }
    //    return 0;
    //}
    compareEntriesByFullValue(a, b) {
        if (a.FullValue < b.FullValue) {
            return -1;
        }
        if (a.FullValue > b.FullValue) {
            return 1;
        }
        return 0;
    }

    //Gets the largest sort index for the entire LookupTable
    //Parameter pwaEntries comes from a JSOM request
    //getLastSortIndex(pwaEntries) {
    //    var largestSortIndex = 0;
    //    const entryEnumerator = pwaEntries.getEnumerator();
    //    while (entryEnumerator.moveNext()) {
    //        let existing = entryEnumerator.get_current();

    //        if (parseFloat(existing.get_sortIndex()) > largestSortIndex) {
    //            console.log(`getLastSortIndex: setting largest index to '${existing.get_sortIndex()}' from fullValue '${existing.get_fullValue()}'`);
    //            largestSortIndex = parseFloat(existing.get_sortIndex());
    //        }
    //    }
    //    return largestSortIndex;
    //}

    //Gets the largest sort index within the chldren of a Parent entry
    //If there are no children it returns the sortIndex of the parent
    //Parameter pwaEntries comes from a JSOM request
    //getLastSortIndexWithinParent(pwaEntries, parentId) {
    //    var entry = this.getEntryById(pwaEntries, parentId);
    //    if (entry == null) {
    //        throw new Error(`Could not find Entry with id=${parentId}`);
    //    }
    //    var parentFullValue = entry.get_fullValue();
    //    console.log(`parentFullValue = '${parentFullValue}'`);
    //    if (parentFullValue === null) {
    //        //The parent is an empty row at the root
    //        console.log(`getLastSortIndexWithinParent: reverting to getLastSortIndex as parent is null`);
    //        return this.getLastSortIndex(pwaEntries);
    //    }
    //    var largestSortIndex = entry.get_sortIndex(); //set default as sortIndex of the parent itself
    //    const entryEnumerator = pwaEntries.getEnumerator();
    //    var fullValue = null;
    //    while (entryEnumerator.moveNext()) {
    //        let existing = entryEnumerator.get_current();
    //        fullValue = existing.get_fullValue();
    //        if (fullValue == null) {
    //            fullValue = ''; //Empty rows can exist within a lookup table
    //        }

    //        if (fullValue.startsWith(parentFullValue)) {
    //            if (parseFloat(existing.get_sortIndex()) > largestSortIndex) {
    //                console.log(`getLastSortIndexWithinParent: setting largest index to '${existing.get_sortIndex()}' from fullValue '${existing.get_fullValue()}'`);
    //                largestSortIndex = parseFloat(existing.get_sortIndex());
    //            }
    //        }
    //    }

    //    return largestSortIndex;
    //}

    //getNextLargestSortIndex(pwaEntries, afterSortIndex) {
    //    const entryEnumerator = pwaEntries.getEnumerator();
    //    afterSortIndex = parseFloat(afterSortIndex);
    //    while (entryEnumerator.moveNext()) {
    //        let existing = entryEnumerator.get_current();
    //        if (afterSortIndex < parseFloat(existing.get_sortIndex())) {
    //            return parseFloat(existing.get_sortIndex());
    //        }
    //    }
    //    return afterSortIndex;
    //}

    //This method should return a sort index that will add a chlid item at the end of the parents children, but before the parents next sibling
    //getNewSortIndexOfChildForParent(pwaEntries, parentId, masks) {

    //    var lastIndexInParent = this.getLastSortIndexWithinParent(pwaEntries, parentId);

    //    var nextSiblingToParent = this.getNextSiblingEntryById(pwaEntries, parentId, masks);
    //    if (nextSiblingToParent == null) {
    //        return (parseFloat(lastIndexInParent) + 1).toFixed(10);
    //    }

    //    lastIndexInParent = parseFloat(lastIndexInParent);
    //    var nextParentSiblingSortIndex = parseFloat(nextSiblingToParent.get_sortIndex());
    //    if (lastIndexInParent == nextParentSiblingSortIndex) {
    //        nextParentSiblingSortIndex = nextParentSiblingSortIndex + 0.5;
    //    }
    //    var newIndex = (lastIndexInParent + nextParentSiblingSortIndex) / 2; //this should give us a fractional value greater than the parent, but less than the first child or next sibling

    //    newIndex = newIndex.toFixed(10);
    //    console.log(`newIndex(${newIndex}) = (lastIndexInParent(${lastIndexInParent}) + nextParentSiblingSortIndex(${nextParentSiblingSortIndex})) / 2`);
    //    return newIndex;

    //}
    //Parameter pwaEntries comes from a JSOM request
    //getEntryById(pwaEntries, id) {
    //    const entryEnumerator = pwaEntries.getEnumerator();
    //    while (entryEnumerator.moveNext()) {
    //        let existing = entryEnumerator.get_current();

    //        if (existing.get_id() == id) {
    //            return existing;
    //        }
    //    }
    //    return null;
    //}
    getEntryByFullValue(entries, fullValue) {
        for (const entry of entries) {
            if (String(entry.FullValue) === String(fullValue)) {
                return entry;
            }
        }
        return null;
    }

    //This method will return next parent that is a sibling to parentId
    //If there is no next parent, it will return null
    //Parameter pwaEntries comes from a JSOM request
    //pwaEntries must already be sorted by sortIndex
    //getNextSiblingEntryById(pwaEntries, id, masks) {
    //    const rootKey = '!!!root!!!'; //Key to identify the root of the look up structure
    //    var useRootKey = false;
    //    var entry = this.getEntryById(pwaEntries, id);
    //    var currentFullValue = entry.get_fullValue();
    //    var originalFullItem = currentFullValue;

    //    if (currentFullValue === null) {
    //        throw new Error(`Entry ${id} has no Value`);
    //    }

    //    var parentFullValue = this.getLookupEntryParentFullValue(currentFullValue, masks);
    //    if (parentFullValue === null) {
    //        parentFullValue = rootKey; //Entry is at root
    //        currentFullValue = rootKey + currentFullValue;
    //        useRootKey = true;
    //    }
    //    var currentFullValueParts = this.getFullValueParts(currentFullValue, masks);
    //    var depthOfItem = currentFullValueParts.length;

    //    const entryEnumerator = pwaEntries.getEnumerator();

    //    var returnNext = false;

    //    while (entryEnumerator.moveNext()) {
    //        let existing = entryEnumerator.get_current();
    //        currentFullValue = existing.get_fullValue();
    //        if (currentFullValue === null) {
    //            currentFullValue = '';
    //        }
    //        if (useRootKey) {
    //            currentFullValue = rootKey + currentFullValue;
    //        }

    //        if (returnNext) {

    //            if (currentFullValue === rootKey && parentFullValue === rootKey) {//Current item is empty and at the root
    //                return existing;
    //            }
    //            if (currentFullValue.startsWith(parentFullValue)) {

    //                currentFullValueParts = this.getFullValueParts(currentFullValue, masks);
    //                var depthOfCurrentItem = currentFullValueParts.length;

    //                if (depthOfCurrentItem == depthOfItem) { //ensure we don't return any of items children

    //                    console.log(`currentFullValue='${currentFullValue}'`);
    //                    console.log(`parentFullValue='${parentFullValue}'`);
    //                    console.log(`originalFullItem='${originalFullItem}'`);

    //                    return existing;
    //                }
    //            }
    //        }
    //        else {
    //            if (existing.get_id() == id) {

    //                if (currentFullValue === rootKey) {
    //                    //Empty row at root, cannot have a parent
    //                    //So return ourselves
    //                    console.log('Empty Row at parent getNextSiblingEntryById is returning itself');
    //                    return existing;
    //                }
    //                else {
    //                    returnNext = true;
    //                }
    //            }
    //        }
    //    }
    //    return null; //There is no next sibling
    //}

}