// Ensure the global SP and PS variables are defined
/// <reference path="PS.debug.js" />
/// <reference path="SP.debug.js" />
/// <reference path="SP.Runtime.js" />
var SP = SP || {};
var PS = PS || {};

class IOResources extends IOBaseEntity {
    constructor(apiUrl, fileInput, logTextArea, progressBar) {
        super(apiUrl, fileInput, logTextArea, progressBar);
    }

    get_name() {
        return "Enterprise Resources";
    }

    async export() {
        this.startNewLog("Retrieving Resources");
        //Using REST
        const data = await this.request('GET', '/_api/ProjectServer/EnterpriseResources');
        //const resources = data.value;
        const resources = data.d.results;

        this.appendLog(`Found ${resources.length} Resources in PWA`);

        //Check which Resources are Users
        let projectContext = PS.ProjectContext.get_current();
        var existingResources = projectContext.get_enterpriseResources();
        projectContext.load(existingResources, 'Include(Name, Id, User)');
        await this.executeQuery(projectContext, () => {
            //All good
        },
        (sender, args) => {
            throw new Error(`FAILED to read Resources from PWA'. Error: '${args.get_message()}'.`);
        });


        //TODO: remove work resources that are users
        for (const resource of resources) {
            if ((resource.ResourceType == 1) && (!resource.IsGeneric)) //Is a Named, Work Resource
            {
                //Non-Generic, Work Resources. These should be users that should be sync'ed via AD sync
                console.log(`Checking if non-generic work resource '${resource.Name}' is also a User.`);
                var existingResource = this.selectResourceByName(existingResources, resource.Name);
                var user = existingResource.get_user();
                projectContext.load(user); //, 'Include(UserId,Email)');
                var isUser = true;
                await this.executeQuery(projectContext, () => {
                    //All good
                    if (user != null) {
                        //Yucky solution, but hey, what can you do with this crap
                        try {
                            console.log(user.get_loginName()); //errors internally if the user object is not an actual user
                        }
                        catch
                        {
                            //console.log('user.get_loginName() = Error');
                            isUser = false;
                        }
                    }
                    else {
                        //console.log('Resource.User is null');
                        isUser = false;
                    }
                },
                (sender, args) => {
                    throw new Error(`FAILED to read User '${resource.Name}' from PWA'. Error: '${args.get_message()}'.`);
                });

                resource.IsUser = isUser; // (user != null); //If user object exists, Resource IS a User
            }
            else {
                resource.IsUser = false;
            }
        }

        //NOTE: there are several properties that are not available via the REST endpoint.
        //It might be better to change this in future to export from the JSOM objects that DO have the properties

        var pwaConfig = {
            EnterpriseResources: resources
        };

        const blob = new Blob([JSON.stringify(pwaConfig, null, 4)], { type: "application/json" });
        const url = URL.createObjectURL(blob);
        const link = document.createElement('a');
        link.href = url;
        link.download = 'Resources.json';
        link.click();
    }

    async import() {
        return new Promise(async (resolve, reject) => {
            try {
                this.startNewLog('Importing Resources');

                const file = this.openFile();
                this.appendLog("Reading Resources from file");
                const fileContent = await this.readFileAsText(file);
                const pwaConfig = JSON.parse(fileContent);
                const resources = pwaConfig.EnterpriseResources;
                if (resources == null) {
                    reject(`No '${this.get_name()}' were found in the selected file.`);
                }

                var count = resources.length;
                this.appendLog(`Found ${count} Resources in file`);
                let projectContext = PS.ProjectContext.get_current();
                this.appendLog("Reading Resources from PWA");
                var existingResources = projectContext.get_enterpriseResources();
                projectContext.load(existingResources, 'Include(Name, Id)');
                await this.executeQuery(projectContext, () => {
                    //All good
                },
                    (sender, args) => {
                        throw new Error(`FAILED to read Resources from PWA'. Error: '${args.get_message()}'.`);
                    });

                this.appendLog(`Found ${existingResources.get_count()} Resources in PWA`);

                var existingResource = null;
                var index = 0;
                var progress = 0;
                var isUpdated = false;
                var isCreated = false;
                for (const resource of resources) {
                    if (!(resource.Selected == null || resource.Selected == true)) {
                        this.appendLog(`Enterprise Resource '${resource.Name}' is not selected for Import.`);
                    }
                    else {
                        existingResource = this.selectResourceByName(existingResources, resource.Name);
                        isUpdated = false;
                        isCreated = false;
                        if (existingResource == null) {
                            if (!resource.IsUser) {
                                try {
                                    isUpdated = true;
                                    this.appendLog(`Resource '${resource.Name}' does not exist, Creating it.`);
                                    var newResource = new PS.EnterpriseResourceCreationInformation();
                                    //newResource.set_id(resource.Id); Causes issue on deleted resources with the same Id
                                    const newGuid = SP.Guid.newGuid().toString();
                                    newResource.set_id(newGuid);

                                    newResource.set_name(resource.Name);
                                    newResource.set_isBudget(resource.IsBudget);
                                    newResource.set_isGeneric(resource.IsGeneric);
                                    newResource.set_isInactive(!resource.IsActive); //Note the ! operator IsInactive vs IsActive
                                    newResource.set_resourceType(resource.ResourceType);

                                    // add more fields as needed, based on your Resources.json data

                                    existingResources.add(newResource);
                                    existingResources.update();
                                    projectContext.load(existingResources);
                                    await this.executeQuery(projectContext, () => {
                                        //Success
                                        //this.appendLog(`Resource '${newResource.get_name()}' has been created.`);
                                    },
                                        (sender, args) => {
                                            throw new Error(`FAILED to create Resource '${newResource.get_name()}'. Error: '${args.get_message()}'.`);
                                        });

                                    //Load new resource
                                    existingResource = this.selectResourceByName(existingResources, resource.Name);
                                    isCreated = true;

                                } catch (error) {
                                    throw new Error(`Error creating resource '${resource.Name}': ${error}`);
                                }
                            }
                            else {
                                this.appendWarning(`Cannot create resource '${resource.Name}' because it is a User. Users must be created by an AD sync or manually.`);

                                index++;
                                progress = Math.round((index / count) * 100);
                                this.updateProgress(progress);
                                continue;//skip to next resource
                            }
                        }

                        projectContext.load(existingResource);
                        await this.executeQuery(projectContext, () => {
                            //Resource loaded
                        },
                            (sender, args) => {
                                throw new Error(`FAILED to load Resource '${existingResource.get_name()}'. Error: '${args.get_message()}'.`);
                            });


                        //Also do Update process even when Resource is just freshly created above, there are lots of fields that are not available on the create process
                        //this.appendLog(`Updating Resource '${resource.Name}'.`);

                        var updatedProperties = this.updateResource(existingResource, resource, isUpdated);
                        if (updatedProperties.length > 0) {
                            isUpdated = true;
                        }

                        if (isUpdated) {
                            existingResources.update();
                            projectContext.load(existingResource);
                            await this.executeQuery(projectContext, () => {
                                if (!isCreated) { //Only report Updated, if not already reported as Created
                                    this.appendLog(`Resource '${existingResource.get_name()}' has been updated. '${String(updatedProperties)}'`);
                                }
                            },
                                (sender, args) => {
                                    throw new Error(`FAILED to update Resource '${existingResource.get_name()}'. Error: '${args.get_message()}'.`);
                                });
                        }
                        else {
                            this.appendLog(`No update required for Resource '${existingResource.get_name()}'.`);
                        }
                    }

                    index++;
                    progress = Math.round((index / count) * 100);
                    this.updateProgress(progress);
                }
                resolve();
            } catch (error) {
                reject(error);
            }
        });
    }

    selectResourceByName(existingResources, resourceName) {
        const resourceEnumerator = existingResources.getEnumerator();
        while (resourceEnumerator.moveNext()) {
            let existing = resourceEnumerator.get_current();

            if (existing.get_name() == resourceName) {
                return existing;
            }
        }
        return null;
    }

    updateResource(existingResource, resource, isUpdated) {
        // Check and update properties if necessary
        var propertiesUpdated = [];
        if (!isUpdated) {
            //These are the create properties that are already set if created new above
            if (existingResource.get_name() != resource.Name) {
                existingResource.set_name(resource.Name);
                propertiesUpdated.push('Name');
            }

            if (existingResource.get_isBudget() != resource.IsBudget) {
                existingResource.set_isBudget(resource.IsBudget);
                propertiesUpdated.push('IsBudget');
            }

            if (existingResource.get_isGeneric() != resource.IsGeneric) {
                existingResource.set_isGeneric(resource.IsGeneric);
                propertiesUpdated.push('IsGeneric');
            }

            if (existingResource.get_isActive() != resource.IsActive) {
                existingResource.set_isActive(resource.IsActive);
                propertiesUpdated.push('IsActive');
            }

            //ResourceType is Read-Only
            //if (existingResource.get_resourceType() != resource.ResourceType) {
            //    existingResource.set_resourceType(resource.ResourceType);
            //    isUpdated = true;
            //}
        }

        //Fields not in REST output are commented in Green
        //existingResource.set_baseCalendar;

        if (existingResource.get_canLevel() != resource.CanLevel) {
            existingResource.set_canLevel(resource.CanLevel);
            propertiesUpdated.push('CanLevel');
        }

        if (existingResource.get_code() != resource.Code) {
            existingResource.set_code(resource.Code);
            propertiesUpdated.push('Code');
        }

        if (existingResource.get_costAccrual() != resource.CostAccrual) {
            existingResource.set_costAccrual(resource.CostAccrual);
            propertiesUpdated.push('CostAccrual');
        }

        if (existingResource.get_costCenter() != resource.CostCenter) {
            existingResource.set_costCenter(resource.CostCenter);
            propertiesUpdated.push('CostCenter');
        }

        //existingResource.set_defaultAssignmentOwner;
        if (existingResource.get_defaultBookingType() != resource.DefaultBookingType) {
            existingResource.set_defaultBookingType(resource.DefaultBookingType);
            propertiesUpdated.push('DefaultBookingType');
        }

        if (existingResource.get_email() != resource.Email) {
            existingResource.set_email(resource.Email);
            propertiesUpdated.push('Email');
        }

        if (existingResource.get_externalId() != resource.ExternalId) {
            existingResource.set_externalId(resource.ExternalId);
            propertiesUpdated.push('ExternalId');
        }

        if (existingResource.get_group() != resource.Group) {
            existingResource.set_group(resource.Group);
            propertiesUpdated.push('Group');
        }

        if (resource.HireDate && resource.HireDate != "0001-01-01T00:00:00") {
            var date = new Date(resource.HireDate);
            if (existingResource.get_hireDate() != date) {
                existingResource.set_hireDate(date);
                propertiesUpdated.push('HireDate');
            }
        }


        if (existingResource.get_initials() != resource.Initials) {
            existingResource.set_initials(resource.Initials);
            propertiesUpdated.push('Initials');
        }

        //isActive, isBudget and isGeneric are covered above

        if (existingResource.get_isTeam() != resource.IsTeam) {
            existingResource.set_isTeam(resource.IsTeam);
            propertiesUpdated.push('IsTeam');
        }

        if (existingResource.get_materialLabel() != resource.MaterialLabel) {
            existingResource.set_materialLabel(resource.MaterialLabel);
            propertiesUpdated.push('MaterialLabel');
        }

        //Name is covered above (but shouldn't need changing anyway)

        if (existingResource.get_phonetics() != resource.Phonetics) {
            existingResource.set_phonetics(resource.Phonetics);
            propertiesUpdated.push('Phonetics');
        }
        //resourceCalendarException will be added as a seperate Entity
        //resourceType is covered above

        if (resource.TerminationDate && resource.TerminationDate != "0001-01-01T00:00:00") {
            var date = new Date(resource.TerminationDate);
            if (existingResource.get_terminationDate() != date) {
                existingResource.set_terminationDate(date);
                propertiesUpdated.push('TerminationDate');
            }
        }

        //existingResource.set_timesheetManager
        //existingResource.set_user //never do this

        //Hyperlink and Notes are note settable properties
        //if (existingResource.get_hyperlink() != resource.Hyperlink) {
        //    existingResource.set_hyperlink(resource.Hyperlink);
        //    isUpdated = true;
        //}

        //if (existingResource.get_hyperlinkAddress() != resource.HyperlinkAddress) {
        //    existingResource.set_hyperlinkAddress(resource.HyperlinkAddress);
        //    isUpdated = true;
        //}

        //if (existingResource.get_notes() != resource.Notes) {
        //    existingResource.set_notes(resource.Notes);
        //    isUpdated = true;
        //}

        return propertiesUpdated;
    }
}