// Ensure the global SP and PS variables are defined
/// <reference path="PS.debug.js" />
/// <reference path="SP.debug.js" />
/// <reference path="SP.Runtime.js" />
var SP = SP || {};
var PS = PS || {};

class IOSiteColumns extends IOBaseEntity {
    constructor(apiUrl, fileInput, logTextArea, progressBar) {
        super(apiUrl, fileInput, logTextArea, progressBar);
    }

    get_name() {
        return "Site Columns";
    }

    async export() {
        this.startNewLog("Retrieving Site Columns");
        //Using REST
        const data = await this.request('GET', '/_api/web/fields');
        //const siteColumns = data.value;
        const siteColumns = data.d.results;

        this.appendLog(`Found ${siteColumns.length} Site Columns in the Site Collection`);
        var pwaConfig = {
            SiteColumns: siteColumns
        };

        const blob = new Blob([JSON.stringify(pwaConfig, null, 4)], { type: "application/json" });
        const url = URL.createObjectURL(blob);
        const link = document.createElement('a');
        link.href = url;
        link.download = 'SiteColumns.json';
        link.click();
    }

    async import() {
        //return new Promise(async (resolve, reject) => {
        //try {
        this.startNewLog('');

        const file = this.openFile();
        this.appendLog("Reading Site Columns from file");
        const fileContent = await this.readFileAsText(file);
        const pwaConfig = JSON.parse(fileContent);
        const siteColumns = pwaConfig.SiteColumns;
        if (siteColumns == null) {
            throw new Error(`No '${this.get_name()}' were found in the selected file.`);
        }

        var count = siteColumns.length;
        this.appendLog(`Found ${count} Site Columns in file`);

        this.appendLog("Reading Site Columns from Site Collection");
        var existingColumns = await this.request('GET', '/_api/web/fields');
        //existingColumns = existingColumns.value;
        existingColumns = existingColumns.d.results;

        this.appendLog(`Found ${existingColumns.length} Site Columns in Site Collection`);

        let projectContext = PS.ProjectContext.get_current();
        let web = projectContext.get_web();
        let fields = projectContext.get_web().get_fields();
        let contentTypes = projectContext.get_web().get_contentTypes();

        projectContext.load(contentTypes, 'Include(Id, Name, StringId, SchemaXml, Fields, FieldLinks)');

        await this.executeQuery(projectContext, async () => {
            //Success
        },
        async (sender, args) => {
            throw new Error(`FAILED to load Site Columns. Error: '${args.get_message()}'.`);
        });

        var builtInSiteColumnsToSkip = this.builtInSiteColumns();
        var index = 0;
        var progress = 0;
        for (const column of siteColumns) {

            if (!(column.Selected == null || column.Selected === true)) {
                this.appendLog(`Site Column '${column.Title} [${column.StaticName}]' is not selected for Import.`);
            }
            else {
                try {
                    let existingColumn = existingColumns.find(c => c.InternalName === column.InternalName);

                    if (existingColumn != null) {
                        // Make updates to the field
                        var importColumnSchemaXML = column.SchemaXml;
                        if (this.versionAttributeExists(importColumnSchemaXML)) {
                            importColumnSchemaXML = this.removeVersionAttribute(importColumnSchemaXML);
                            //this.appendWarning(`Removed 'Version' attribute from Site Column '${column.Title}'`);
                        }
                        //else {
                        //    this.appendWarning(`Site Column '${column.Title}' has no 'Version' attribute to remove`);
                        //}

                        if (existingColumn.SchemaXml === importColumnSchemaXML) {
                            this.appendLog(`Site Column '${column.Title}' already exists and doesn't require updates. Skipping.`);
                            //No log entry required, makes it too big
                        }
                        else {
                            if (builtInSiteColumnsToSkip.includes(column.Title)) {
                                this.appendWarning(`Skipping Site Column '${column.Title}' because it is a special BuiltIn Site Column`);
                                continue;
                            }

                            let fieldToUpdate = fields.getByInternalNameOrTitle(existingColumn.InternalName);
                            projectContext.load(fieldToUpdate);

                            await this.executeQuery(projectContext, async () => {
                                //Success
                            },
                            async (sender, args) => {
                                throw new Error(`FAILED to load Site Column '${column.Title}' for update. Error: '${args.get_message()}'.`);
                            });

                            fieldToUpdate.set_schemaXml(importColumnSchemaXML);

                            //fieldToUpdate.update();
                            fieldToUpdate.updateAndPushChanges(true);

                            var tryDeleteRecreate = false;
                            try {
                                await this.executeQuery(projectContext, () => {
                                    this.appendLog(`Updated Site Column '${column.Title}' and pushed changes to lists.`);
                                },
                                //error received
                                (sender, args) => {
                                    if (args.get_message().includes('The object has been updated by another user since it was last fetched')) {
                                        tryDeleteRecreate = true;
                                    }
                                    throw new Error(`FAILED to update Site Column '${column.Title}'. Error: '${args.get_message()}'.`);
                                });
                            }
                            catch (err)
                            {
                                const urlParams = new URLSearchParams(window.location.search);
                                var allowDelete = urlParams.has('AllowDelete');

                                //Had an issue where SchemaXML included a "Version" attribute. This blocked Site Columns from 
                                //being updated in future changes. This error handler resolves the issue by trapping the resulting error
                                //Deleteing the Site Column, after removing it from Content Types, and then re adding it back as new
                                //After running this, the content Types that used the column will be broken, so Content Types need to be re-imported.
                                //Handle the Site Column Version fix here
                                if (allowDelete && tryDeleteRecreate) {
                                    this.appendWarning(`Deleting and recreate Site Column '${column.Title}'`);

                                    await this.deleteSiteColumn(projectContext, web, fieldToUpdate, column.Title, contentTypes);
                                    fields = await this.createSiteColumn(projectContext, fields, importColumnSchemaXML, column);

                                    this.appendWarning(`Don't forget to re-import Content Types now, as this column '${column.Title}' will be missing from them.`);
                                }
                                else {
                                    this.appendWarning('AllowDelete is disabled, throwing error');
                                    this.appendError(err.message);
                                }
                            }
                        }
                    }
                    else
                    {
                        if (builtInSiteColumnsToSkip.includes(column.Title)) {
                            this.appendWarning(`Skipping Site Column '${column.Title}' because it is a special BuiltIn Site Column`);
                            continue;
                        }

                        this.appendLog(`Site Column '${column.Title}' does not exist. Creating it.`);

                        var importColumnSchemaXML = column.SchemaXml;
                        if (this.versionAttributeExists(importColumnSchemaXML)) {
                            importColumnSchemaXML = this.removeVersionAttribute(importColumnSchemaXML);
                        }

                        fields = await this.createSiteColumn(projectContext, fields, importColumnSchemaXML, column);
                    }
                }
                catch (error) {
                    this.appendError(error.toString()); //Log error and move on to next Site Column
                }
            }
            index++;
            progress = Math.round((index / count) * 100); //leave 1 % for the actual save
            this.updateProgress(progress);
        }

        //this.updateProgress(100);

        //resolve();
        //}
        //catch (error) {
        //this.appendWarning(`${error.toString()}, rowNumber = ${rowNumber}, index = ${index}`);
        //reject(error);
        //}
        //});
    }

    builtInSiteColumns() {
        var builtInSiteColumnsToSkip = [];
        builtInSiteColumnsToSkip.push("Enterprise Keywords"); //added by Title
        builtInSiteColumnsToSkip.push("Wiki Categories"); //Fails with "Uncaught TypeError: Cannot read properties of undefined (reading 'TaxonomyField')"

        return builtInSiteColumnsToSkip;
    }

    versionAttributeExists(schemaXml) {
        var parser = new DOMParser();
        var xmlSchemaDoc = parser.parseFromString(schemaXml, "text/xml");
        if (xmlSchemaDoc.getElementsByTagName("Field") != null) {
            var fieldElement = xmlSchemaDoc.getElementsByTagName("Field")[0];
            if (fieldElement != null) {
                return fieldElement.hasAttribute("Version");
            }
        }
        return false;
    }

    removeVersionAttribute(schemaXml) {
        var parser = new DOMParser();
        var xmlSchemaDoc = parser.parseFromString(schemaXml, "text/xml");
        if (xmlSchemaDoc.getElementsByTagName("Field") != null) {
            var fieldElement = xmlSchemaDoc.getElementsByTagName("Field")[0];
            if (fieldElement != null) {
                if (fieldElement.hasAttribute("Version")) {
                    fieldElement.removeAttribute("Version");
                    var serializedXml = new XMLSerializer().serializeToString(fieldElement);
                    return serializedXml;
                }
            }
        }
        return schemaXml;
    }

    async createSiteColumn(projectContext, fields, schemaXML, column) {
        // Use column.SchemaXml to add field to fields collection
        fields.addFieldAsXml(schemaXML, false, SP.AddFieldOptions.addToNoContentType);
        
        await this.executeQuery(projectContext, async () => {
            this.appendLog(`Site Column '${column.Title}' has been created.`);
        },
        async (sender, args) => {
            var msg = args.get_message();
            //if (msg === "The object has been updated by another user since it was last fetched.") {
            //    this.appendLog(`Site Column '${column.Title}' has been created.`); //we always get this error but it creates anyway.
            //    //this.appendWarning(`We "think" Site Column '${column.Title}' has been created.`);
            //    //this.appendWarning("An error was received, but its a red herring error that we found indicates the Site Column existed once before but was deleted and has now been re-created anyway.");
            //}
            //else {
            this.appendAndThrowError(`FAILED to create Site Column '${column.Title}'. Error: '${args.get_message()}'.`);

            //}
        });

        //re-load fields collection
        fields = projectContext.get_web().get_fields();
        await this.executeQuery(projectContext, async () => {
            //Success
        },
        async (sender, args) => {
            throw new Error(`FAILED to reload Site Columns. Error: '${args.get_message()}'.`);
        });

        return fields;
    }

    async deleteSiteColumn(projectContext, web, fieldToDelete, fieldTitle, contentTypes) {

        //First we find all Content Types the Site Column is referenced in and build a list of positions in them
        //var usedInContentTypes = [];

        const contentTypeEnumerator = contentTypes.getEnumerator();
        while (contentTypeEnumerator.moveNext()) {
            let contentType = contentTypeEnumerator.get_current();

            const fieldLinkEnumerator = contentType.get_fieldLinks().getEnumerator();
            while (fieldLinkEnumerator.moveNext()) {
                let fieldLink = fieldLinkEnumerator.get_current();

                if (fieldLink.get_id().toString() === fieldToDelete.get_id().toString()) { //For some reason the id's don't equate when they should, maybe they are of different types. Converting to strings for compare checking
                    fieldLink.deleteObject();
                    contentType.update(false); //Do not update inheriting content types

                    await this.executeQuery(projectContext, () => {
                        this.appendWarning(`Removed Site Column '${fieldTitle}' from Content Type '${contentType.get_name()}'`);
                    },
                    (sender, args) => {
                        this.appendAndThrowError(`FAILED to remove Site Column '${fieldTitle}' from Content Type '${contentType.get_name()}'. Error: '${args.get_message()}'.`);
                    });
                    break;
                }
            }
        }

        //Reload fieldToDelete to exclude CT references it currently holds
        projectContext.load(fieldToDelete);
        projectContext.load(web);

        fieldToDelete.deleteObject();
        web.update();

        await this.executeQuery(projectContext, () => {
            this.appendWarning(`Deleted Site Column '${fieldTitle}'`);
        },
        (sender, args) => {
            this.appendAndThrowError(`FAILED to delete Site Column '${fieldTitle}'. Error: '${args.get_message()}'.`);
        });
    }
}